"""
This module provides a `DateTimeColumn` class that extends the `Column` class from PySpark. It allows for adding or
subtracting an interval value from a datetime column.

This can be used to reflect a change in a given date / time column in a more human-readable way.

Please refer to the Spark SQL documentation for a list of valid interval values:
https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html#interval-literal

Background
----------
The aim is to easily add or subtract an 'interval' value to a datetime column. An interval value is a string that
represents a time interval. For example, '1 day', '1 month', '5 years', '1 minute 30 seconds', '10 milliseconds', etc.
These can be used to reflect a change in a given date / time column in a more human-readable way.

Typically, this can be done using the `date_add()` and `date_sub()` functions in Spark SQL. However, these functions
only support adding or subtracting a single unit of time measured in days. Using an interval gives us much more
flexibility; however, Spark SQL does not provide a function to add or subtract an interval value from a datetime column
through the python API directly, so we have to use the `expr()` function to do this to be able to directly use SQL.

This module provides a `DateTimeColumn` class that extends the `Column` class from PySpark. It allows for adding or
subtracting an interval value from a datetime column using the `+` and `-` operators.

Additionally, this module provides two transformation classes that can be used as a transformation step in a pipeline:

- `DateTimeAddInterval`: adds an interval value to a datetime column
- `DateTimeSubtractInterval`: subtracts an interval value from a datetime column

These classes are subclasses of `ColumnsTransformationWithTarget` and hence can be used to perform transformations on
multiple columns at once.

The above transformations both use the provided `asjust_time()` function to perform the actual transformation.

See also:
---------
Related Koheesio classes:

[ColumnsTransformation]: ../index.md#koheesio.spark.transformations.ColumnsTransformation
[ColumnsTransformationWithTarget]: ../index.md#koheesio.spark.transformations.ColumnsTransformationWithTarget

From the koheesio.spark.transformations module:

* [ColumnsTransformation] : Base class for ColumnsTransformation. Defines column / columns field + recursive logic
* [ColumnsTransformationWithTarget] : Defines target_column / target_suffix field

pyspark.sql.functions:

-   https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html#interval-literal
-   https://spark.apache.org/docs/latest/api/sql/index.html
-   https://spark.apache.org/docs/latest/api/sql/#try_add
-   https://spark.apache.org/docs/latest/api/sql/#try_subtract

Classes
-------

DateTimeColumn
    A datetime column that can be adjusted by adding or subtracting an interval value using the `+` and `-` operators.

DateTimeAddInterval
    A transformation that adds an interval value to a datetime column.
    This class is a subclass of `ColumnsTransformationWithTarget` and hence can be used as a transformation step in a
    pipeline. See `ColumnsTransformationWithTarget` for more information.

DateTimeSubtractInterval
    A transformation that subtracts an interval value from a datetime column.
    This class is a subclass of `ColumnsTransformationWithTarget` and hence can be used as a transformation step in a
    pipeline. See `ColumnsTransformationWithTarget` for more information.

Note
----
the `DateTimeAddInterval` and `DateTimeSubtractInterval` classes are very similar. The only difference is that one
adds an interval value to a datetime column, while the other subtracts an interval value from a datetime column.

Functions
---------
dt_column(column: Union[str, Column]) -> DateTimeColumn
    Converts a column to a `DateTimeColumn`. This function aims to be a drop-in replacement for
    `pyspark.sql.functions.col` that returns a `DateTimeColumn` instead of a `Column`.

adjust_time(column: Column, operation: Operations, interval: str) -> Column
    Adjusts a datetime column by adding or subtracting an interval value.

validate_interval(interval: str)
    Validates a given interval string.

Example
-------
#### Various ways to create and interact with `DateTimeColumn`:
- Create a `DateTimeColumn` from a string: `dt_column("my_column")`
- Create a `DateTimeColumn` from a `Column`: `dt_column(df.my_column)`
- Use the `+` and `-` operators to add or subtract an interval value from a `DateTimeColumn`:
    - `dt_column("my_column") + "1 day"`
    - `dt_column("my_column") - "1 month"`

#### Functional examples using `adjust_time()`:
- Add 1 day to a column: `adjust_time("my_column", operation="add", interval="1 day")`
- Subtract 1 month from a column: `adjust_time("my_column", operation="subtract", interval="1 month")`

#### As a transformation step:
```python
from koheesio.spark.transformations.date_time.interval import (
    DateTimeAddInterval,
)

input_df = spark.createDataFrame(
    [(1, "2022-01-01 00:00:00")], ["id", "my_column"]
)

# add 1 day to my_column and store the result in a new column called 'one_day_later'
output_df = DateTimeAddInterval(
    column="my_column", target_column="one_day_later", interval="1 day"
).transform(input_df)
```
__output_df__:

| id | my_column           | one_day_later       |
|----|---------------------|---------------------|
| 1  | 2022-01-01 00:00:00 | 2022-01-02 00:00:00 |

`DateTimeSubtractInterval` works in a similar way, but subtracts an interval value from a datetime column.
"""

from typing import Literal, Union

from pyspark.sql import Column
from pyspark.sql.functions import col, expr
from pyspark.sql.utils import ParseException

from koheesio.models import Field, field_validator
from koheesio.spark.transformations import ColumnsTransformationWithTarget

# create a literal constraining the operations to 'add' and 'subtract'
Operations = Literal["add", "subtract"]


class DateTimeColumn(Column):
    """A datetime column that can be adjusted by adding or subtracting an interval value  using the `+` and `-`
    operators.
    """

    def __add__(self, value: str):
        """Add an `interval` value to a date or time column

        A valid value is a string that can be parsed by the `interval` function in Spark SQL.
        See https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html#interval-literal
        """
        return self.from_column(adjust_time(self, operation="add", interval=value))

    def __sub__(self, value: str):
        """Subtract an `interval` value to a date or time column

        A valid value is a string that can be parsed by the `interval` function in Spark SQL.
        See https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html#interval-literal
        """
        return self.from_column(adjust_time(self, operation="subtract", interval=value))

    @classmethod
    def from_column(cls, column: Column):
        """Create a DateTimeColumn from an existing Column"""
        return cls(column._jc)


def validate_interval(interval: str):
    """Validate an interval string

    Parameters
    ----------
    interval : str
        The interval string to validate

    Raises
    ------
    ValueError
        If the interval string is invalid
    """
    try:
        expr(f"interval '{interval}'")
    except ParseException as e:
        raise ValueError(f"Value '{interval}' is not a valid interval.") from e
    return interval


def dt_column(column: Union[str, Column]) -> DateTimeColumn:
    """Convert a column to a DateTimeColumn

    Aims to be a drop-in replacement for `pyspark.sql.functions.col` that returns a DateTimeColumn instead of a Column.

    Example
    --------
    ### create a DateTimeColumn from a string
    ```python
    dt_column("my_column")
    ```

    ### create a DateTimeColumn from a Column
    ```python
    dt_column(df.my_column)
    ```

    Parameters
    ----------
    column : Union[str, Column]
        The column (or name of the column) to convert to a DateTimeColumn
    """
    if isinstance(column, str):
        column = col(column)
    elif not isinstance(column, Column):
        raise TypeError(f"Expected column to be of type str or Column, got {type(column)} instead.")
    return DateTimeColumn.from_column(column)


def adjust_time(column: Column, operation: Operations, interval: str) -> Column:
    """
    Adjusts a datetime column by adding or subtracting an interval value.

    This can be used to reflect a change in a given date / time column in a more human-readable way.


    See also
    --------
    Please refer to the Spark SQL documentation for a list of valid interval values:
    https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html#interval-literal

    ### pyspark.sql.functions:

    * https://spark.apache.org/docs/latest/api/sql/index.html#interval
    * https://spark.apache.org/docs/latest/api/sql/#try_add
    * https://spark.apache.org/docs/latest/api/sql/#try_subtract

    Example
    --------
    ### add 1 day to a column
    ```python
    adjust_time("my_column", operation="add", interval="1 day")
    ```

    ### subtract 1 month from a column
    ```python
    adjust_time("my_column", operation="subtract", interval="1 month")
    ```

    ### or, a much more complicated example

    In this example, we add 5 days, 3 hours, 7 minutes, 30 seconds, and 1 millisecond to a column called `my_column`.
    ```python
    adjust_time(
        "my_column",
        operation="add",
        interval="5 days 3 hours 7 minutes 30 seconds 1 millisecond",
    )
    ```

    Parameters
    ----------
    column : Column
        The datetime column to adjust.
    operation : Operations
        The operation to perform. Must be either 'add' or 'subtract'.
    interval : str
        The value to add or subtract. Must be a valid interval string.

    Returns
    -------
    Column
        The adjusted datetime column.
    """

    # check that value is a valid interval
    interval = validate_interval(interval)

    column_name = column._jc.toString()

    # determine the operation to perform
    try:
        operation = {
            "add": "try_add",
            "subtract": "try_subtract",
        }[operation]
    except KeyError as e:
        raise ValueError(f"Operation '{operation}' is not valid. Must be either 'add' or 'subtract'.") from e

    # perform the operation
    _expression = f"{operation}({column_name}, interval '{interval}')"
    column = expr(_expression)

    return column


class DateTimeAddInterval(ColumnsTransformationWithTarget):
    """
    A transformation that adds or subtracts a specified interval from a datetime column.

    See also:
    ---------
    pyspark.sql.functions:

    -   https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html#interval-literal
    -   https://spark.apache.org/docs/latest/api/sql/index.html#interval

    Parameters
    ----------
    interval : str
        The interval to add to the datetime column.
    operation : Operations, optional, default=add
        The operation to perform. Must be either 'add' or 'subtract'.

    Example
    -------
    ### add 1 day to a column
    ```python
    DateTimeAddInterval(
        column="my_column",
        interval="1 day",
    ).transform(df)
    ```

    ### subtract 1 month from `my_column` and store the result in a new column called `one_month_earlier`
    ```python
    DateTimeSubtractInterval(
        column="my_column",
        target_column="one_month_earlier",
        interval="1 month",
    )
    ```
    """

    interval: str = Field(
        default=...,
        description="The interval to add to the datetime column.",
        examples=["1 day", "5 years", "3 months"],
    )
    operation: Operations = Field(
        default="add", description="The operation to perform. Must be either 'add' or 'subtract'."
    )

    # validators
    validate_interval = field_validator("interval")(validate_interval)

    def func(self, column: Column):
        return adjust_time(column, operation=self.operation, interval=self.interval)


class DateTimeSubtractInterval(DateTimeAddInterval):
    """Subtracts a specified interval from a datetime column.

    Works in the same way as `DateTimeAddInterval`, but subtracts the specified interval from the datetime column.
    See `DateTimeAddInterval` for more information.
    """

    operation: Operations = Field(
        default="subtract", description="The operation to perform. Must be either 'add' or 'subtract'."
    )



import logging
from io import StringIO

import pytest
from requests_mock.mocker import Mocker

from pydantic import SecretStr

from koheesio.sso import okta as o


class TestOktaToken:
    url = "https://host.okta.com/oauth2/auth/v1/token"
    ot = o.OktaAccessToken(
        url=url,
        client_id="client",
        client_secret=SecretStr("secret"),
    )

    def test_okta_token_w_token(self):
        with Mocker() as rm:
            rm.post(self.url, json={"access_token": "bar"}, status_code=int(200))
            assert self.ot.execute().token.get_secret_value() == "bar"

    def test_okta_token_wo_token(self):
        with Mocker() as rm:
            rm.post(self.url, json={"foo": "bar"}, status_code=int(200))
            with pytest.raises(ValueError):
                self.ot.execute()

    def test_okta_token_non_200(self):
        with Mocker() as rm:
            rm.post(self.url, status_code=int(404))
            with pytest.raises(o.HTTPError):
                self.ot.execute()

    def test_wo_extra_params(self):
        oat = o.OktaAccessToken(url="url", client_id="client", client_secret=SecretStr("secret"))
        assert oat.params == {"auth": ("client", "secret")}

    def test_w_extra_params(self):
        oat = o.OktaAccessToken(url="url", client_id="client", client_secret=SecretStr("secret"), params={"foo": "bar"})
        assert oat.params == {"foo": "bar", "auth": ("client", "secret")}

    def test_log_extra_params_secret(self):
        log_capture_string = StringIO()
        ch = logging.StreamHandler(log_capture_string)
        ch.setLevel(logging.DEBUG)
        logger = logging.getLogger("tests")
        logger.addHandler(ch)
        secret_val = "secret_value"
        oat = o.OktaAccessToken(
            url="url", client_id="client", client_secret=SecretStr(secret_val), params={"foo": "bar"}
        )
        logger.warning(f"{oat.params = }")
        log_contents = log_capture_string.getvalue()
        assert "secret" not in log_contents
        assert "*" * len(secret_val) + "(Masked)" in log_contents



"""
Spark step module
"""

from __future__ import annotations

from typing import Optional
from abc import ABC

from pydantic import Field

from pyspark.sql import Column
from pyspark.sql import DataFrame as PySparkSQLDataFrame
from pyspark.sql import SparkSession as OriginalSparkSession
from pyspark.sql import functions as F

try:
    from pyspark.sql.utils import AnalysisException as SparkAnalysisException
except ImportError:
    from pyspark.errors.exceptions.base import AnalysisException as SparkAnalysisException

from koheesio import Step, StepOutput
from koheesio.models import model_validator

# TODO: Move to spark/__init__.py after reorganizing the code
# Will be used for typing checks and consistency, specifically for PySpark >=3.5
DataFrame = PySparkSQLDataFrame
SparkSession = OriginalSparkSession
AnalysisException = SparkAnalysisException


class SparkStep(Step, ABC):
    """Base class for a Spark step

    Extends the Step class with SparkSession support. The following:
    - Spark steps are expected to return a Spark DataFrame as output.
    - spark property is available to access the active SparkSession instance.
    - The SparkSession instance can be provided as an argument to the constructor through the `spark` parameter.
    """

    spark: Optional[SparkSession] = Field(
        default=None,
        description="The SparkSession instance. If not provided, the active SparkSession will be used.",
        validate_default=False,
    )

    class Output(StepOutput):
        """Output class for SparkStep"""

        df: Optional[DataFrame] = Field(default=None, description="The Spark DataFrame")

    @model_validator(mode="after")
    def _get_active_spark_session(self):
        """Return active SparkSession instance
        If a user provides a SparkSession instance, it will be returned. Otherwise, an active SparkSession will be
        attempted to be retrieved.
        """
        if self.spark is None:
            self.spark = SparkSession.getActiveSession()
        return self


# TODO: Move to spark/functions/__init__.py after reorganizing the code
def current_timestamp_utc(spark: SparkSession) -> Column:
    """Get the current timestamp in UTC"""
    return F.to_utc_timestamp(F.current_timestamp(), spark.conf.get("spark.sql.session.timeZone"))



"""Transform module

Transform aims to provide an easy interface for calling transformations on a Spark DataFrame, where the transformation
is a function that accepts a DataFrame (df) and any number of keyword args.
"""

from __future__ import annotations

from typing import Callable, Dict
from functools import partial

from pyspark.sql import DataFrame

from koheesio.models import ExtraParamsMixin, Field
from koheesio.spark.transformations import Transformation
from koheesio.utils import get_args_for_func


class Transform(Transformation, ExtraParamsMixin):
    """
    Transform aims to provide an easy interface for calling transformations on a Spark DataFrame,
    where the transformation is a function that accepts a DataFrame (df) and any number of keyword args.

    The implementation is inspired by and based upon:
    https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrame.transform.html

    Parameters
    ----------
    func : Callable
        The function to be called on the DataFrame.
    params : Dict, optional, default=None
        The keyword arguments to be passed to the function. Defaults to None. Alternatively, keyword arguments can be
        passed directly as keyword arguments - they will be merged with the `params` dictionary.

    Example
    -------
    ### a function compatible with Transform:
    ```python
    def some_func(df, a: str, b: str):
        return df.withColumn(a, f.lit(b))
    ```

    ### verbose style input in Transform
    ```python
    Transform(func=some_func, params={"a": "foo", "b": "bar"})
    ```

    ### shortened style notation (easier to read)
    ```python
    Transform(some_func, a="foo", b="bar")
    ```

    ### when too much input is given, Transform will ignore extra input
    ```python
    Transform(
        some_func,
        a="foo",
        # ignored input
        c="baz",
        title=42,
        author="Adams",
        # order of params input should not matter
        b="bar",
    )
    ```

    ### using the from_func classmethod
    ```python
    SomeFunc = Transform.from_func(some_func, a="foo")
    some_func = SomeFunc(b="bar")
    ```
    """

    func: Callable = Field(default=None, description="The function to be called on the DataFrame.")

    def __init__(self, func: Callable, params: Dict = None, df: DataFrame = None, **kwargs):
        params = {**(params or {}), **kwargs}
        super().__init__(func=func, params=params, df=df)

    def execute(self):
        """Call the function on the DataFrame with the given keyword arguments."""
        func, kwargs = get_args_for_func(self.func, self.params)
        self.output.df = self.df.transform(func=func, **kwargs)

    @classmethod
    def from_func(cls, func: Callable, **kwargs) -> Callable[..., Transform]:
        """Create a Transform class from a function. Useful for creating a new class with a different name.

        This method uses the `functools.partial` function to create a new class with the given function and keyword
        arguments. This way you can pre-define some of the keyword arguments for the function that might be needed for
        the specific use case.

        Example
        -------
        ```python
        CustomTransform = Transform.from_func(some_func, a="foo")
        some_func = CustomTransform(b="bar")
        ```

        In this example, `CustomTransform` is a Transform class with the function `some_func` and the keyword argument
        `a` set to "foo". When calling `some_func(b="bar")`, the function `some_func` will be called with the keyword
        arguments `a="foo"` and `b="bar"`.
        """
        return partial(cls, func=func, **kwargs)



"""
Test the Trim columns transformation
"""

import pytest

from pydantic import ValidationError

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.strings.trim import LTrim, RTrim, Trim

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name=__file__, inherit_from_koheesio=True)


@pytest.mark.parametrize(
    "input_values,input_data,input_schema,expected",
    [
        (
            # description: Test Trim
            # input values
            dict(
                columns=["_1"],
                target_column="result",
            ),
            # input data
            [[" ltext"], ["rtext "], [" lrtext "]],
            # input schema
            ["_1"],
            # expected data
            dict(
                Trim=["ltext", "rtext", "lrtext"],
                LTrim=["ltext", "rtext ", "lrtext "],
                RTrim=[" ltext", "rtext", " lrtext"],
            ),
        )
    ],
)
def test_happy_flow(input_values, input_data, input_schema, expected, spark):
    input_df = spark.createDataFrame(input_data, input_schema)

    klz = Trim, LTrim, RTrim

    for kls in klz:
        log.info(f"testing for {kls.__name__}")
        trim = kls(**input_values)
        output_df = trim.transform(input_df)
        target_column = trim.target_column

        # log equivalent of doing df.show()
        log.info(f"show output_df: \n{output_df._jdf.showString(20, 20, False)}")
        actual = [row[target_column] for row in output_df.select(target_column).collect()]
        assert actual == expected[kls.__name__]


@pytest.mark.parametrize(
    "input_values,expected",
    [
        (
            # case 1 : no column specified -> trim all string columns
            dict(),
            # expected output
            [
                dict(id=1, input_column="foo", other_column="buz"),
                dict(id=2, input_column="bar", other_column=None),
            ],
        ),
        (
            # case 2 : string and integer column specified -> only one string column affected
            dict(column=["input_column", "id"]),
            # expected output
            [
                dict(id=1, input_column="foo", other_column="  buz  "),
                dict(id=2, input_column="bar", other_column=None),
            ],
        ),
    ],
)
def test_base(input_values, expected, spark):
    input_df = spark.createDataFrame(
        [[1, "foo  ", "  buz  "], [2, "  bar  ", None]], ["id", "input_column", "other_column"]
    )

    df = Trim(**input_values).transform(input_df)

    actual = [k.asDict() for k in df.collect()]
    assert actual == expected


def test_invalid_direction():
    with pytest.raises(ValidationError):
        Trim(direction="invalid")


# TODO: not supported yet
#
#     def test_trim_nested(self):
#         df = self.spark.createDataFrame([{"a": {"b": "    hello world   "}}], self.schema)
#         df = (
#             Trim(
#                 df=df,
#                 params={"source_column": "a.b"},
#                 target_alias="a.c",
#                 columns_to_flatten=["a{}.b"],
#             )
#             .execute()
#             .df
#         )
#         result = df.rdd.map(lambda r: r.a).collect()
#         self.assertEqual(result[0]["c"], "hello world")
#
#     def test_field_does_not_exist(self):
#         with self.assertRaises(AnalysisException) as e:
#             Trim(
#                 df=self.sample_df,
#                 params={"source_column": "some_column"},
#                 target_alias="a",
#             ).execute()
#
#         self.assertIn(
#             "cannot resolve 'some_column' given input columns:",
#             e.exception.desc.replace("'`", "'").replace("`'", "'"),
#         )



"""Dummy step for testing purposes.

This module contains a dummy step for testing purposes. It is used to test the Koheesio framework or to provide a simple
example of how to create a new step.

Example
-------
```python
s = DummyStep(a="a", b=2)
s.execute()
```
In this case, `s.output` will be equivalent to the following dictionary:
```python
{"a": "a", "b": 2, "c": "aa"}
```
"""

from koheesio import Step, StepOutput


class DummyOutput(StepOutput):
    """Dummy output for testing purposes."""

    a: str
    b: int


class DummyStep(Step):
    """Dummy step for testing purposes."""

    a: str
    b: int

    class Output(DummyOutput):
        """Dummy output for testing purposes."""

        c: str

    def execute(self):
        """Dummy execute for testing purposes."""
        self.output.a = self.a
        self.output.b = self.b
        self.output.c = self.a * self.b



"""
The first test, tests that OktaAccessToken can be instantiated without pyspark
This test should not raise an ImportError

The second test, tests that a sparksession needs pyspark to be available
This test should raise an ImportError hence pytest.raises(ImportError) is expected

"""

from unittest import mock

import pytest

from pyspark.sql import SparkSession

from koheesio.models import SecretStr
from koheesio.spark import SparkStep

pytestmark = pytest.mark.spark


class TestSparkImportFailures:
    def test_import_error_no_error(self):
        secret = SecretStr("client_secret")
        with mock.patch.dict("sys.modules", {"pyspark": None}):
            from koheesio.sso.okta import OktaAccessToken

            OktaAccessToken(url="https://nike.okta.com", client_id="client_id", client_secret=secret)

    def test_import_error_with_error(self):
        with mock.patch.dict("sys.modules", {"pyspark.sql": None, "koheesio.steps.spark": None}):
            with pytest.raises(ImportError):
                from pyspark.sql import SparkSession

                SparkSession.builder.appName("tests").getOrCreate()

            pass


class TestSparkStep:
    """Test SparkStep class"""

    def test_spark_property_with_session(self):
        spark = SparkSession.builder.appName("pytest-pyspark-local-testing-explicit").master("local[*]").getOrCreate()
        step = SparkStep(spark=spark)
        assert step.spark is spark

    def test_spark_property_without_session(self):
        spark = SparkSession.builder.appName("pytest-pyspark-local-testing-implicit").master("local[*]").getOrCreate()
        step = SparkStep()
        assert step.spark is spark



"""
This module defines the DeltaTableWriter class, which is used to write both batch and streaming dataframes
to Delta tables.

DeltaTableWriter supports two output modes: `MERGEALL` and `MERGE`.

- The `MERGEALL` mode merges all incoming data with existing data in the table based on certain conditions.
- The `MERGE` mode allows for more custom merging behavior using the DeltaMergeBuilder class from the `delta.tables`
  library.

The `output_mode_params` dictionary is used to specify conditions for merging, updating, and inserting data.
The `target_alias` and `source_alias` keys are used to specify the aliases for the target and source dataframes in the
merge conditions.

Classes
-------
DeltaTableWriter
    A class for writing data to Delta tables.
DeltaTableStreamWriter
    A class for writing streaming data to Delta tables.

Example
-------
```python
DeltaTableWriter(
    table="test_table",
    output_mode=BatchOutputMode.MERGEALL,
    output_mode_params={
        "merge_cond": "target.id=source.id",
        "update_cond": "target.col1_val>=source.col1_val",
        "insert_cond": "source.col_bk IS NOT NULL",
    },
)
```
"""

from typing import List, Optional, Set, Type, Union
from functools import partial

from delta.tables import DeltaMergeBuilder, DeltaTable
from py4j.protocol import Py4JError

from pyspark.sql import DataFrameWriter

from koheesio.models import ExtraParamsMixin, Field, field_validator
from koheesio.spark.delta import DeltaTableStep
from koheesio.spark.utils import on_databricks
from koheesio.spark.writers import BatchOutputMode, StreamingOutputMode, Writer
from koheesio.spark.writers.delta.utils import log_clauses


class DeltaTableWriter(Writer, ExtraParamsMixin):
    """Delta table Writer for both batch and streaming dataframes.

    Example
    -------
    ### Example for `MERGEALL`
    ```python
    DeltaTableWriter(
        table="test_table",
        output_mode=BatchOutputMode.MERGEALL,
        output_mode_params={
            "merge_cond": "target.id=source.id",
            "update_cond": "target.col1_val>=source.col1_val",
            "insert_cond": "source.col_bk IS NOT NULL",
            "target_alias": "target",  # <------ DEFAULT, can be changed by providing custom value
            "source_alias": "source",  # <------ DEFAULT, can be changed by providing custom value
        },
    )
    ```

    ### Example for `MERGE`
    ```python
    DeltaTableWriter(
        table="test_table",
        output_mode=BatchOutputMode.MERGE,
        output_mode_params={
            'merge_builder': (
                DeltaTable
                .forName(sparkSession=spark, tableOrViewName=<target_table_name>)
                .alias(target_alias)
                .merge(source=df, condition=merge_cond)
                .whenMatchedUpdateAll(condition=update_cond)
                .whenNotMatchedInsertAll(condition=insert_cond)
                )
            }
        )
    ```

    ### Example for `MERGE`
    in case the table isn't created yet, first run will execute an APPEND operation
    ```python
    DeltaTableWriter(
        table="test_table",
        output_mode=BatchOutputMode.MERGE,
        output_mode_params={
            "merge_builder": [
                {
                    "clause": "whenMatchedUpdate",
                    "set": {"value": "source.value"},
                    "condition": "<update_condition>",
                },
                {
                    "clause": "whenNotMatchedInsert",
                    "values": {"id": "source.id", "value": "source.value"},
                    "condition": "<insert_condition>",
                },
            ],
            "merge_cond": "<merge_condition>",
        },
    )
    ```

    ### Example for APPEND
    dataframe writer options can be passed as keyword arguments
    ```python
    DeltaTableWriter(
        table="test_table",
        output_mode=BatchOutputMode.APPEND,
        partitionOverwriteMode="dynamic",
        mergeSchema="false",
    )
    ```

    Parameters
    ----------
    table : Union[DeltaTableStep, str]
        The table to write to
    output_mode : Optional[Union[str, BatchOutputMode, steps.writers.StreamingOutputMode]]
        The output mode to use. Default is BatchOutputMode.APPEND. For streaming, use StreamingOutputMode.
    params : Optional[dict]
        Additional parameters to use for specific mode
    """

    table: Union[DeltaTableStep, str] = Field(default=..., description="The table to write to")
    output_mode: Optional[Union[BatchOutputMode, StreamingOutputMode]] = Field(
        default=BatchOutputMode.APPEND,
        alias="outputMode",
        description=f"{BatchOutputMode.__doc__}\n{StreamingOutputMode.__doc__}",
    )
    params: Optional[dict] = Field(
        default_factory=dict,
        alias="output_mode_params",
        description="Additional parameters to use for specific mode",
    )

    partition_by: Optional[List[str]] = Field(
        default=None, alias="partitionBy", description="The list of fields to partition the Delta table on"
    )
    format: str = "delta"  # The format to use for writing the dataframe to the Delta table

    _merge_builder: DeltaMergeBuilder = None

    # noinspection PyProtectedMember
    def __merge(self, merge_builder: Optional[DeltaMergeBuilder] = None) -> Union[DeltaMergeBuilder, DataFrameWriter]:
        """Merge dataframes using DeltaMergeBuilder or DataFrameWriter"""
        builder, writer = None, None

        if self.table.exists:
            merge_builder = self._get_merge_builder(merge_builder)

            if on_databricks():
                try:
                    source_alias = merge_builder._jbuilder.getMergePlan().source().alias()
                    target_alias = merge_builder._jbuilder.getMergePlan().target().alias()
                    self.log.debug(
                        f"The following aliases are used during Merge operation: source={source_alias}, target={target_alias}"
                    )
                    patched__log_clauses = partial(log_clauses, source_alias=source_alias, target_alias=target_alias)
                    self.log.debug(
                        patched__log_clauses(clauses=merge_builder._jbuilder.getMergePlan().matchedClauses())
                    )
                    self.log.debug(
                        patched__log_clauses(clauses=merge_builder._jbuilder.getMergePlan().notMatchedClauses())
                    )
                    self.log.debug(
                        patched__log_clauses(clauses=merge_builder._jbuilder.getMergePlan().notMatchedBySourceClauses())
                    )
                except Py4JError:
                    self.log.debug("Merge plan has not been received, skipping logging of merge details")

            builder = merge_builder

        else:
            writer = self.df

            if self.output_mode == BatchOutputMode.MERGEALL.value:
                writer = writer.alias(self.params.get("source_alias", "source"))
                insert_cond = self.params.get("insert_cond")
                filter_cond = insert_cond if insert_cond else "1=1"
                self.log.info(
                    f"Table `{self.table.table_name}` doesn't exist. Using `append` instead of `{self.output_mode}`."
                    f"Data will be filtered using filter from whenNotMatchedInsertAll: `{filter_cond}`"
                )
                writer = writer.filter(filter_cond)
            elif self.output_mode == BatchOutputMode.MERGE.value:
                self.log.info(
                    f"Table `{self.table.table_name}` doesn't exist. Using `append` instead of `{self.output_mode}`."
                    f"Insert filter from DeltaMergeBuilder will NOT be applied."
                    f"As workaround, create Delta table beforehand."
                )
            else:
                raise ValueError(f"Unsupported output_mode `{self.output_mode}` for merge operation.")

            writer = writer.write.format("delta").mode("append")

        return builder or writer

    def __merge_all(self) -> Union[DeltaMergeBuilder, DataFrameWriter]:
        """Merge dataframes using DeltaMergeBuilder or DataFrameWriter"""
        merge_cond = self.params.get("merge_cond", None)

        if merge_cond is None:
            raise ValueError(
                "Provide `merge_cond` in DeltaTableWriter(output_mode_params={'merge_cond':'<str or Column>'})"
            )

        update_cond = self.params.get("update_cond", None)
        insert_cond = self.params.get("insert_cond", None)
        target_alias = self.params.get("target_alias", "target")
        source_alias = self.params.get("target_alias", "source")

        if self.table.exists:
            builder = (
                DeltaTable.forName(sparkSession=self.spark, tableOrViewName=self.table.table_name)
                .alias(target_alias)
                .merge(source=self.df.alias(source_alias), condition=merge_cond)
                .whenMatchedUpdateAll(condition=update_cond)
                .whenNotMatchedInsertAll(condition=insert_cond)
            )
        else:
            builder = None

        return self.__merge(merge_builder=builder)

    def _get_merge_builder(self, provided_merge_builder=None) -> DeltaMergeBuilder:
        """Resolves the merge builder. If provided, it will be used, otherwise it will be created from the args"""

        # A merge builder has been already created - case for merge_all
        if provided_merge_builder:
            self._merge_builder = provided_merge_builder
            return provided_merge_builder

        if self._merge_builder:
            return self._merge_builder

        merge_builder = self.params.get("merge_builder", None)

        # Merge builder can be built in two ways: directly from the Delta table (if it already exists), or from the
        # provided args
        if merge_builder:
            if isinstance(merge_builder, DeltaMergeBuilder):
                return merge_builder

            if isinstance(merge_builder, list) and "merge_cond" in self.params:  # type: ignore
                return self._merge_builder_from_args()

        raise ValueError(
            "Provide `merge_builder` in "
            "DeltaTableWriter(output_mode_params={'merge_builder': DeltaMergeBuilder()}). "
            "See documentation for options."
        )

    def _merge_builder_from_args(self):
        """Creates the DeltaMergeBuilder from the provided configuration"""
        merge_clauses = self.params.get("merge_builder", None)
        merge_cond = self.params.get("merge_cond", None)
        source_alias = self.params.get("source_alias", "source")
        target_alias = self.params.get("target_alias", "target")

        builder = (
            DeltaTable.forName(self.spark, self.table.table_name)
            .alias(target_alias)
            .merge(self.df.alias(source_alias), merge_cond)
        )

        for merge_clause in merge_clauses:
            clause_type = merge_clause.pop("clause", None)
            method = getattr(builder, clause_type)
            builder = method(**merge_clause)

        return builder

    @field_validator("output_mode")
    def _validate_output_mode(cls, mode):
        """Validate `output_mode` value"""
        if isinstance(mode, str):
            mode = cls.get_output_mode(mode, options={StreamingOutputMode, BatchOutputMode})
        if not isinstance(mode, BatchOutputMode) and not isinstance(mode, StreamingOutputMode):
            raise AttributeError(
                f"""
                Invalid outputMode specified '{mode}'. Allowed values are:
                Batch Mode - {BatchOutputMode.__doc__}
                Streaming Mode - {StreamingOutputMode.__doc__}
                """
            )
        return str(mode.value)

    @field_validator("table")
    def _validate_table(cls, table):
        """Validate `table` value"""
        if isinstance(table, str):
            return DeltaTableStep(table=table)
        return table

    @field_validator("params")
    def _validate_params(cls, params):
        """Validates params. If an array of merge clauses is provided, they will be validated against the available
        ones in DeltaMergeBuilder"""

        valid_clauses = {m for m in dir(DeltaMergeBuilder) if m.startswith("when")}

        if "merge_builder" in params:
            merge_builder = params["merge_builder"]
            if isinstance(merge_builder, list):
                for merge_conf in merge_builder:
                    clause = merge_conf.get("clause")
                    if clause not in valid_clauses:
                        raise ValueError(f"Invalid merge clause '{clause}' provided")
            elif not isinstance(merge_builder, DeltaMergeBuilder):
                raise ValueError("merge_builder must be a list or merge clauses or a DeltaMergeBuilder instance")

        return params

    @classmethod
    def get_output_mode(cls, choice: str, options: Set[Type]) -> Union[BatchOutputMode, StreamingOutputMode]:
        """Retrieve an OutputMode by validating `choice` against a set of option OutputModes.

        Currently supported output modes can be found in:

        - BatchOutputMode
        - StreamingOutputMode
        """
        for enum_type in options:
            if choice.upper() in [om.value.upper() for om in enum_type]:
                return getattr(enum_type, choice.upper())
        raise AttributeError(
            f"""
            Invalid outputMode specified '{choice}'. Allowed values are:
            Batch Mode - {BatchOutputMode.__doc__}
            Streaming Mode - {StreamingOutputMode.__doc__}
            """
        )

    def __data_frame_writer(self) -> DataFrameWriter:
        """Specify DataFrameWriter for batch mode"""
        _writer = self.df.write.format(self.format)
        if self.partition_by:
            _writer = _writer.partitionBy(self.partition_by)
        return _writer.mode(self.output_mode)

    @property
    def writer(self) -> Union[DeltaMergeBuilder, DataFrameWriter]:
        """Specify DeltaTableWriter"""
        map_mode_writer = {
            BatchOutputMode.MERGEALL.value: self.__merge_all,
            BatchOutputMode.MERGE.value: self.__merge,
        }

        return map_mode_writer.get(self.output_mode, self.__data_frame_writer)()

    def execute(self):
        _writer = self.writer

        if self.table.create_if_not_exists and not self.table.exists:
            _writer = _writer.options(**self.table.default_create_properties)

        if isinstance(_writer, DeltaMergeBuilder):
            _writer.execute()
        else:
            if options := self.params:
                # should we add options only if mode is not merge?
                _writer = _writer.options(**options)
            _writer.saveAsTable(self.table.table_name)



"""
Pad the values of a column with a character up until it reaches a certain length.

Classes
-------
Pad
    Pads the values of `source_column` with the `character` up until it reaches `length` of characters

LPad
    Pad with a character on the left side of the string.

RPad
    Pad with a character on the right side of the string.
"""

from typing import Literal, Optional

from pyspark.sql import Column
from pyspark.sql.functions import lpad, rpad

from koheesio.models import Field, PositiveInt, constr
from koheesio.spark.transformations import ColumnsTransformationWithTarget

pad_directions = Literal["left", "right"]


class Pad(ColumnsTransformationWithTarget):
    """
    Pads the values of `source_column` with the `character` up until it reaches `length` of characters
    The `direction` param can be changed to apply either a left or a right pad. Defaults to left pad.

    Wraps the `lpad` and `rpad` functions from PySpark.

    Parameters
    ----------
    columns : Union[str, List[str]]
        The column (or list of columns) to pad. Alias: column
    target_column : Optional[str], optional, default=None
        The column to store the result in. If not provided, the result will be stored in the source column.
        Alias: target_suffix - if multiple columns are given as source, this will be used as a suffix.
    character : constr(min_length=1)
        The character to use for padding
    length : PositiveInt
        Positive integer to indicate the intended length
    direction : Optional[pad_directions], optional, default=left
        On which side to add the characters. Either "left" or "right". Defaults to "left"

    Example
    -------
    __input_df:__

    | column |
    |--------|
    | hello  |
    | world  |

    ```python
    output_df = Pad(
        column="column",
        target_column="padded_column",
        character="*",
        length=10,
        direction="right",
    ).transform(input_df)
    ```

    __output_df:__

    | column | padded_column |
    |--------|---------------|
    | hello  | hello*****    |
    | world  | world*****    |

    > Note: in the example above, we could have used the RPad class instead of Pad with direction="right" to achieve the
    same result.

    """

    character: constr(min_length=1) = Field(default=..., description="The character to use for padding")
    length: PositiveInt = Field(default=..., description="Positive integer to indicate the intended length")
    direction: Optional[pad_directions] = Field(
        default="left", description='On which side to add the characters . Either "left" or "right". Defaults to "left"'
    )

    def func(self, column: Column):
        func = lpad if self.direction == "left" else rpad
        return func(column, self.length, self.character)


LPad = Pad  # Pad's default is already 'left', so LPad is just an alias of Pad


class RPad(Pad):
    """Pad with a character on the right side of the string.

    See Pad class docstring for more information.
    """

    direction: Optional[pad_directions] = "right"



from unittest import mock

import pytest

from pyspark.sql import SparkSession

from koheesio.spark.readers.jdbc import JdbcReader

pytestmark = pytest.mark.spark


class TestJdbcReader:
    common_options = {
        "driver": "driver",
        "url": "url",
        "user": "user",
        "password": "password",
    }

    def test_get_options_wo_extra_options(self):
        jr = JdbcReader(**self.common_options)
        actual = jr.get_options()
        del actual["password"]  # we don't need to test for this

        expected = {**self.common_options}
        del expected["password"]  # we don't need to test for this

        assert actual == expected

    def test_get_options_w_extra_options(self):
        jr = JdbcReader(
            options={
                "foo": "foo",
                "bar": "bar",
            },
            **self.common_options,
        )

        actual = jr.get_options()
        del actual["password"]  # we don't need to test for this

        expected = {
            **self.common_options,
            "foo": "foo",
            "bar": "bar",
        }
        del expected["password"]  # we don't need to test for this

        assert actual == expected

    def test_execute_wo_dbtable_and_query(self):
        jr = JdbcReader(**self.common_options)
        with pytest.raises(ValueError) as e:
            jr.execute()
            assert e.type is ValueError

    def test_execute_w_dbtable_and_query(self, dummy_spark):
        with mock.patch.object(SparkSession, "getActiveSession") as mock_spark:
            mock_spark.return_value = dummy_spark

            jr = JdbcReader(**self.common_options, dbtable="foo", query="bar")
            jr.execute()

            assert jr.df.count() == 1
            assert mock_spark.return_value.options_dict["query"] == "bar"
            assert "dbtable" not in mock_spark.return_value.options_dict

    def test_execute_w_dbtable(self, dummy_spark):
        with mock.patch.object(SparkSession, "getActiveSession") as mock_spark:
            mock_spark.return_value = dummy_spark

            jr = JdbcReader(**self.common_options, dbtable="foo")
            jr.execute()

            assert jr.df.count() == 1
            assert mock_spark.return_value.options_dict["dbtable"] == "foo"



"""
HANA reader.
"""

from typing import Any, Dict, Optional

from koheesio.models import Field
from koheesio.spark.readers.jdbc import JdbcReader


class HanaReader(JdbcReader):
    """
    Wrapper around JdbcReader for SAP HANA

    Notes
    -----
    * Refer to [JdbcReader](jdbc.md#koheesio.spark.readers.jdbc.JdbcReader) for the list of all available parameters.
    * Refer to SAP HANA Client Interface Programming Reference docs for the list of all available connection string
        parameters:
        https://help.sap.com/docs/SAP_HANA_CLIENT/f1b440ded6144a54ada97ff95dac7adf/109397c2206a4ab2a5386d494f4cf75e.html

    Example
    -------
    > Note: jars should be added to the Spark session manually. This class does not take care of that.

    This example depends on the SAP HANA `ngdbc` JAR. e.g. ngdbc-2.5.49.

    ```python
    from koheesio.spark.readers.hana import HanaReader
    jdbc_hana = HanaReader(
        url="jdbc:sap://<domain_or_ip>:<port>/?<options>
        user="YOUR_USERNAME",
        password="***",
        dbtable="schemaname.tablename"
    )
    df = jdbc_hana.read()
    ```

    Parameters
    ----------
    url : str
        JDBC connection string. Refer to SAP HANA docs for the list of all available connection string parameters.
        Example: jdbc:sap://<server>:<port>[/?<options>]
    user : str
    password : SecretStr
    dbtable : str
        Database table name, also include schema name
    options : Optional[Dict[str, Any]]
        Extra options to pass to the SAP HANA JDBC driver. Refer to SAP HANA docs for the list of all available
        connection string parameters. Example: {"fetchsize": 2000, "numPartitions": 10}
    query : Optional[str]
        Query
    format : str
        The type of format to load. Defaults to 'jdbc'. Should not be changed.
    driver : str
        Driver name. Be aware that the driver jar needs to be passed to the task. Should not be changed.
    """

    driver: str = Field(
        default="com.sap.db.jdbc.Driver",
        description="Make sure that the necessary JARs are available in the cluster: ngdbc-2-x.x.x.x",
    )
    options: Optional[Dict[str, Any]] = Field(
        default={"fetchsize": 2000, "numPartitions": 10},
        description="Extra options to pass to the SAP HANA JDBC driver",
    )



"""
This module provides utility functions while working with delta framework.
"""

from typing import Optional

from py4j.java_gateway import JavaObject


def log_clauses(clauses: JavaObject, source_alias: str, target_alias: str) -> Optional[str]:
    """
    Prepare log message for clauses of DeltaMergePlan statement.

    Parameters
    ----------
    clauses : JavaObject
        The clauses of the DeltaMergePlan statement.
    source_alias : str
        The source alias.
    target_alias : str
        The target alias.

    Returns
    -------
    Optional[str]
        The log message if there are clauses, otherwise None.

    Notes
    -----
    This function prepares a log message for the clauses of a DeltaMergePlan statement. It iterates over the clauses,
    processes the conditions, and constructs the log message based on the clause type and columns.

    If the condition is a value, it replaces the source and target aliases in the condition string. If the condition is
    None, it sets the condition_clause to "No conditions required".

    The log message includes the clauses type, the clause type, the columns, and the condition.
    """
    log_message = None

    if not clauses.isEmpty():
        clauses_type = clauses.last().nodeName().replace("DeltaMergeInto", "")
        _processed_clauses = {}

        for i in range(0, clauses.length()):
            clause = clauses.apply(i)
            condition = clause.condition()

            if "value" in dir(condition):
                condition_clause = (
                    condition.value()
                    .toString()
                    .replace(f"'{source_alias}", source_alias)
                    .replace(f"'{target_alias}", target_alias)
                )
            elif condition.toString() == "None":
                condition_clause = "No conditions required"

            clause_type: str = clause.clauseType().capitalize()
            columns = "ALL" if clause_type == "Delete" else clause.actions().toList().apply(0).toString()

            if clause_type.lower() not in _processed_clauses:
                _processed_clauses[clause_type.lower()] = []

            log_message = (
                f"{clauses_type} will perform action:{clause_type} columns ({columns}) if `{condition_clause}`"
            )

    return log_message



"""
Extracts a substring from a string column starting at the given position.
"""

from typing import Optional

from pyspark.sql import Column
from pyspark.sql.functions import substring, when
from pyspark.sql.types import StringType

from koheesio.models import Field, PositiveInt, field_validator
from koheesio.spark.transformations import ColumnsTransformationWithTarget


class Substring(ColumnsTransformationWithTarget):
    """
    Extracts a substring from a string column starting at the given position.

    This is a wrapper around PySpark substring() function

    Notes
    -----
    - Numeric columns will be cast to string
    - start is 1-indexed, not 0-indexed!

    Parameters
    ----------
    columns : Union[str, List[str]]
        The column (or list of columns) to substring. Alias: column
    target_column : Optional[str], optional, default=None
        The column to store the result in. If not provided, the result will be stored in the source column.
        Alias: target_suffix - if multiple columns are given as source, this will be used as a suffix.
    start : PositiveInt
        Positive int. Defines where to begin the substring from. The first character of the field has index 1!
    length : Optional[int], optional, default=-1
        Optional. If not provided, the substring will go until end of string.

    Example
    -------
    ### Extract a substring from a string column starting at the given position.

    __input_df:__

    | column    |
    |-----------|
    | skyscraper|

    ```python
    output_df = Substring(
        column="column",
        target_column="substring_column",
        start=3,  # 1-indexed! So this will start at the 3rd character
        length=4,
    ).transform(input_df)
    ```

    __output_df:__

    | column    | substring_column |
    |-----------|------------------|
    | skyscraper| yscr             |

    """

    start: PositiveInt = Field(default=..., description="The starting position")
    length: Optional[int] = Field(
        default=-1,
        description="The target length for the string. use -1 to perform until end",
    )

    @field_validator("length")
    def _valid_length(cls, length_value):
        """Integer.maxint fix for Java.
        Python's sys.maxsize is larger which makes f.substring fail"""
        if length_value == -1:
            return 2147483647
        return length_value

    def func(self, column: Column):
        return when(column.isNull(), None).otherwise(substring(column, self.start, self.length)).cast(StringType())
